#!/usr/bin/env python3
import os
import dill
import unittest
import ratings

TEST_DIRECTORY = os.path.dirname(__file__)


def compare_lists(x, y, test_name, db_name):
    assert len(x) == len(
        y), f'Failure while testing {test_name} with {db_name} db:\n Expected list of length {len(x)} but got {len(y)}'
    assert isinstance(x, type(
        y)), f'Failure while testing {test_name} with {db_name} db:\n Expected type {type(x)} but got {type(y)}'
    x2 = sorted(x)
    y2 = sorted(y)
    for exp, res in zip(x2, y2):
        emsg = f'Failure while testing {test_name} with {db_name} db:\n Element at index {y.index(res)}: {res} not found in expected result'
        if len(x2) < 10:
            emsg += f'exp={x2}\nres={y2}'
        assert exp == res or abs(exp-res) <= 1e-6, emsg


def compare_sets(x, y, test_name, db_name):
    assert len(x) == len(
        y), f'Failure while testing {test_name} with {db_name} db:\n Expected set of length {len(x)} but got {len(y)}'
    assert isinstance(x, type(
        y)), f'Failure while testing {test_name} with {db_name} db:\n Expected type {type(x)} but got {type(y)}'
    for item in y:
        assert item in x, f'Failure while testing {test_name} with {db_name} db:\n Item {item} not found in expected result'


def compare_dictionary(x, y, test_name, db_name):
    assert len(x) == len(
        y), f'Failure while testing {test_name} with {db_name} db:\n Expected {len(x)} number of keys but got {len(y)}'
    compare_sets(set(x.keys()), set(y.keys()), test_name, db_name)

    for key in y:
        result_values = y[key]
        expected = x[key]
        if isinstance(expected, list):
            try:
                compare_lists(expected, result_values, test_name, db_name)
            except BaseException:
                assert False, f'Failure while testing {test_name} with {db_name} db:\n Expected key {key} to have value {expected} but got {result_values}\n (Note the order of the values in the list do not matter)'
        if isinstance(expected, set):
            try:
                compare_sets(expected, result_values, test_name, db_name)
            except BaseException:
                assert False, f'Failure while testing {test_name} with {db_name} db:\n Expected key {key} to have value {expected} but got {result_values}\n (Note the order of the values in the set do not matter)'

class TestUserRatings(unittest.TestCase):
    def test_make_user_ratings(self):
        test_name = 'make_user_ratings'
        for db_name, db in test_results['db'].items():
            expected = test_results[test_name][db_name]
            result = ratings.make_user_ratings(db)
            compare_dictionary(expected, result, test_name, db_name)
            self.assertEqual(test_results['db'][db_name], test_results_copy['db'][db_name])


class TestItemRatings(unittest.TestCase):
    def test_make_item_ratings(self):
        test_name = 'make_item_ratings'
        for db_name, db in test_results['db'].items():
            expected = test_results[test_name][db_name]
            result = ratings.make_item_ratings(db)
            compare_dictionary(expected, result, test_name, db_name)
            self.assertEqual(test_results['db'][db_name], test_results_copy['db'][db_name])


class TestFilter(unittest.TestCase):
    def test_tiny_filter(self):
        tiny_reviews = test_results['db']['tiny']
        item_ratings = ratings.make_item_ratings(tiny_reviews)
        sm_filt = lambda r, k: len(r[k]) > 2
        exp_sm = {'desk': [4.0, 4.0, 5.0, 5.0], 'stapler': [1.0, 5.0, 1.0]}
        compare_dictionary(exp_sm, ratings.filter_ratings(item_ratings, sm_filt), 'tiny_item_ratings', 'filtered')
        # no item in tiny has 5 or more reviews!
        self.assertEqual(ratings.filter_ratings(item_ratings, ratings.filter5), {})

        medium_reviews = test_results['db']['medium']
        item_ratings = ratings.make_item_ratings(medium_reviews)
        res = ratings.filter_ratings(item_ratings, ratings.filter5)
        self.assertEqual(len(res), 142)


    def test_filter_ratings(self):
        test_name = 'filter_ratings'
        for db_name, _ in test_results['db'].items():
            expected = test_results[test_name][db_name]
            for filt, rating_dict, exp in expected:
                rcopy = {k:v[:] for k,v in rating_dict.items()}
                result = ratings.filter_ratings(rating_dict, filt)
                compare_dictionary(exp, result, f'{test_name}_{filt.__name__}', db_name)
                self.assertEqual(rating_dict, rcopy)
                #"Make sure that filter_ratings doesn't modify input!"

class TestBest(unittest.TestCase):
    def test_tiny_best(self):
        tiny_reviews = test_results['db']['tiny']
        item_ratings = ratings.make_item_ratings(tiny_reviews)
        self.assertEqual(ratings.larger_num_ratings(item_ratings, "desk", "stapler"), "desk")
        self.assertEqual(ratings.larger_num_ratings(item_ratings, "rubber bands", "stapler"), "stapler")
        self.assertEqual(ratings.larger_num_ratings(item_ratings, "rubber bands", "desk"), "desk")
        self.assertEqual(ratings.find_best(item_ratings, ratings.larger_num_ratings), "desk")

        medium_reviews = test_results['db']['medium']
        item_ratings = ratings.make_item_ratings(medium_reviews)
        res = ratings.find_best(item_ratings, ratings.larger_num_ratings)
        exp = "B00AE9V3WQ"
        self.assertEqual(res, exp)

    def test_find_best(self):
        test_name = 'find_best'
        for db_name, _ in test_results['db'].items():
            expected = test_results[test_name][db_name]
            for comp, rating_dict, exp in expected:
                rcopy = {k:v[:] for k,v in rating_dict.items()}
                result = ratings.find_best(rating_dict, comp)
                self.assertEqual(exp, result)
                self.assertEqual(rating_dict, rcopy)


if __name__ == '__main__':
    import sys

    thismodule = sys.modules[__name__]

    filename = os.path.join(
        TEST_DIRECTORY,
        'resources',
        'tests',
        f'db_tests.pickle')
    with open(filename, 'rb') as f:
        raw = dill.load(f)
        for db_name, db in raw['db'].items():
            setattr(thismodule, f'db_{db_name}', db)
        setattr(thismodule, 'test_results', raw)
        setattr(thismodule, 'test_results_copy', raw)

    unittest.main()
    print("done testing!")